﻿using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Threading;

namespace BeepMap
{
    public partial class Main : Form
    {
        sndDirectSound ds;
        WaveGen waveGen;
        bool unsavedChanges = false;
        string activeFile = "Untitled.beep";
        string activePath = "";

        ArrayList channels = new ArrayList();
        ArrayList waveforms = new ArrayList();
        int chanCount = 0, currentChannel = 0, currentPattern = 0;
        bool deleting = false, initializing = false, newChanInit = true;

        /*public const float BaseC = 16.35f;
        public const float BaseCSharp = 17.32f;
        public const float BaseD = 18.35f;
        public const float BaseDSharp = 19.45f;
        public const float BaseE = 20.60f;
        public const float BaseF = 21.83f;
        public const float BaseFSharp = 23.12f;
        public const float BaseG = 24.50f;
        public const float BaseGSharp = 25.96f;
        public const float BaseA = 27.50f;
        public const float BaseASharp = 29.14f;
        public const float BaseB = 30.87f;*/

        public Main()
        {
            InitializeComponent();

            // Initialize NoteOffsetMap

            /*
             * This is extremely hacky and inefficient, but it is 5:03 A.M. and
             * I really wanted to get this functionality implemented.
             * 
             * You can observe its inefficiency by the lag in hovering over keys
             * in the piano conrol when running BeepMap.
             * 
             * Using hashtable to be lazy and use negative integers as keys <= Hack
             * Using dictionary and adding 69 to each key >= Probably better
             * Some other solution? == Best?
            */
            PianoKey.NoteOffsetMap.Add(-69, "C-1");
            PianoKey.NoteOffsetMap.Add(-68, "C#-1");
            PianoKey.NoteOffsetMap.Add(-67, "D-1");
            PianoKey.NoteOffsetMap.Add(-66, "D#-1");
            PianoKey.NoteOffsetMap.Add(-65, "E-1");
            PianoKey.NoteOffsetMap.Add(-64, "F-1");
            PianoKey.NoteOffsetMap.Add(-63, "F#-1");
            PianoKey.NoteOffsetMap.Add(-62, "G-1");
            PianoKey.NoteOffsetMap.Add(-61, "G#-1");
            PianoKey.NoteOffsetMap.Add(-60, "A-1");
            PianoKey.NoteOffsetMap.Add(-59, "A#-1");
            PianoKey.NoteOffsetMap.Add(-58, "B-1");

            PianoKey.NoteOffsetMap.Add(-57, "C0");
            PianoKey.NoteOffsetMap.Add(-56, "C#0");
            PianoKey.NoteOffsetMap.Add(-55, "D0");
            PianoKey.NoteOffsetMap.Add(-54, "D#0");
            PianoKey.NoteOffsetMap.Add(-53, "E0");
            PianoKey.NoteOffsetMap.Add(-52, "F0");
            PianoKey.NoteOffsetMap.Add(-51, "F#0");
            PianoKey.NoteOffsetMap.Add(-50, "G0");
            PianoKey.NoteOffsetMap.Add(-49, "G#0");
            PianoKey.NoteOffsetMap.Add(-48, "A0");
            PianoKey.NoteOffsetMap.Add(-47, "A#0");
            PianoKey.NoteOffsetMap.Add(-46, "B0");

            PianoKey.NoteOffsetMap.Add(-45, "C1");
            PianoKey.NoteOffsetMap.Add(-44, "C#1");
            PianoKey.NoteOffsetMap.Add(-43, "D1");
            PianoKey.NoteOffsetMap.Add(-42, "D#1");
            PianoKey.NoteOffsetMap.Add(-41, "E1");
            PianoKey.NoteOffsetMap.Add(-40, "F1");
            PianoKey.NoteOffsetMap.Add(-39, "F#1");
            PianoKey.NoteOffsetMap.Add(-38, "G1");
            PianoKey.NoteOffsetMap.Add(-37, "G#1");
            PianoKey.NoteOffsetMap.Add(-36, "A1");
            PianoKey.NoteOffsetMap.Add(-35, "A#1");
            PianoKey.NoteOffsetMap.Add(-34, "B1");

            PianoKey.NoteOffsetMap.Add(-33, "C2");
            PianoKey.NoteOffsetMap.Add(-32, "C#2");
            PianoKey.NoteOffsetMap.Add(-31, "D2");
            PianoKey.NoteOffsetMap.Add(-30, "D#2");
            PianoKey.NoteOffsetMap.Add(-29, "E2");
            PianoKey.NoteOffsetMap.Add(-28, "F2");
            PianoKey.NoteOffsetMap.Add(-27, "F#2");
            PianoKey.NoteOffsetMap.Add(-26, "G2");
            PianoKey.NoteOffsetMap.Add(-25, "G#2");
            PianoKey.NoteOffsetMap.Add(-24, "A2");
            PianoKey.NoteOffsetMap.Add(-23, "A#2");
            PianoKey.NoteOffsetMap.Add(-22, "B2");

            PianoKey.NoteOffsetMap.Add(-21, "C3");
            PianoKey.NoteOffsetMap.Add(-20, "C#3");
            PianoKey.NoteOffsetMap.Add(-19, "D3");
            PianoKey.NoteOffsetMap.Add(-18, "D#3");
            PianoKey.NoteOffsetMap.Add(-17, "E3");
            PianoKey.NoteOffsetMap.Add(-16, "F3");
            PianoKey.NoteOffsetMap.Add(-15, "F#3");
            PianoKey.NoteOffsetMap.Add(-14, "G3");
            PianoKey.NoteOffsetMap.Add(-13, "G#3");
            PianoKey.NoteOffsetMap.Add(-12, "A3");
            PianoKey.NoteOffsetMap.Add(-11, "A#3");
            PianoKey.NoteOffsetMap.Add(-10, "B3");

            PianoKey.NoteOffsetMap.Add(-9, "C4");
            PianoKey.NoteOffsetMap.Add(-8, "C#4");
            PianoKey.NoteOffsetMap.Add(-7, "D4");
            PianoKey.NoteOffsetMap.Add(-6, "D#4");
            PianoKey.NoteOffsetMap.Add(-5, "E4");
            PianoKey.NoteOffsetMap.Add(-4, "F4");
            PianoKey.NoteOffsetMap.Add(-3, "F#4");
            PianoKey.NoteOffsetMap.Add(-2, "G4");
            PianoKey.NoteOffsetMap.Add(-1, "G#4");
            PianoKey.NoteOffsetMap.Add(0, "A4");
            PianoKey.NoteOffsetMap.Add(1, "A#4");
            PianoKey.NoteOffsetMap.Add(2, "B4");

            PianoKey.NoteOffsetMap.Add(3, "C5");
            PianoKey.NoteOffsetMap.Add(4, "C#5");
            PianoKey.NoteOffsetMap.Add(5, "D5");
            PianoKey.NoteOffsetMap.Add(6, "D#5");
            PianoKey.NoteOffsetMap.Add(7, "E5");
            PianoKey.NoteOffsetMap.Add(8, "F5");
            PianoKey.NoteOffsetMap.Add(9, "F#5");
            PianoKey.NoteOffsetMap.Add(10, "G5");
            PianoKey.NoteOffsetMap.Add(11, "G#5");
            PianoKey.NoteOffsetMap.Add(12, "A5");
            PianoKey.NoteOffsetMap.Add(13, "A#5");
            PianoKey.NoteOffsetMap.Add(14, "B5");

            PianoKey.NoteOffsetMap.Add(15, "C6");
            PianoKey.NoteOffsetMap.Add(16, "C#6");
            PianoKey.NoteOffsetMap.Add(17, "D6");
            PianoKey.NoteOffsetMap.Add(18, "D#6");
            PianoKey.NoteOffsetMap.Add(19, "E6");
            PianoKey.NoteOffsetMap.Add(20, "F6");
            PianoKey.NoteOffsetMap.Add(21, "F#6");
            PianoKey.NoteOffsetMap.Add(22, "G6");
            PianoKey.NoteOffsetMap.Add(23, "G#6");
            PianoKey.NoteOffsetMap.Add(24, "A6");
            PianoKey.NoteOffsetMap.Add(25, "A#6");
            PianoKey.NoteOffsetMap.Add(26, "B6");

            PianoKey.NoteOffsetMap.Add(27, "C7");
            PianoKey.NoteOffsetMap.Add(28, "C#7");
            PianoKey.NoteOffsetMap.Add(29, "D7");
            PianoKey.NoteOffsetMap.Add(30, "D#7");
            PianoKey.NoteOffsetMap.Add(31, "E7");
            PianoKey.NoteOffsetMap.Add(32, "F7");
            PianoKey.NoteOffsetMap.Add(33, "F#7");
            PianoKey.NoteOffsetMap.Add(34, "G7");
            PianoKey.NoteOffsetMap.Add(35, "G#7");
            PianoKey.NoteOffsetMap.Add(36, "A7");
            PianoKey.NoteOffsetMap.Add(37, "A#7");
            PianoKey.NoteOffsetMap.Add(38, "B7");

            PianoKey.NoteOffsetMap.Add(39, "C7");
            PianoKey.NoteOffsetMap.Add(40, "C#7");
            PianoKey.NoteOffsetMap.Add(41, "D7");
            PianoKey.NoteOffsetMap.Add(42, "D#7");
            PianoKey.NoteOffsetMap.Add(43, "E7");
            PianoKey.NoteOffsetMap.Add(44, "F7");
            PianoKey.NoteOffsetMap.Add(45, "F#7");
            PianoKey.NoteOffsetMap.Add(46, "G7");
            PianoKey.NoteOffsetMap.Add(47, "G#7");
            PianoKey.NoteOffsetMap.Add(48, "A7");
            PianoKey.NoteOffsetMap.Add(49, "A#7");
            PianoKey.NoteOffsetMap.Add(50, "B7");

            PianoKey.NoteOffsetMap.Add(51, "C8");
            PianoKey.NoteOffsetMap.Add(52, "C#8");
            PianoKey.NoteOffsetMap.Add(53, "D8");
            PianoKey.NoteOffsetMap.Add(54, "D#8");
            PianoKey.NoteOffsetMap.Add(55, "E8");
            PianoKey.NoteOffsetMap.Add(56, "F8");
            PianoKey.NoteOffsetMap.Add(57, "F#8");
            PianoKey.NoteOffsetMap.Add(58, "G8");
            PianoKey.NoteOffsetMap.Add(59, "G#8");
            PianoKey.NoteOffsetMap.Add(60, "A8");
            PianoKey.NoteOffsetMap.Add(61, "A#8");
            PianoKey.NoteOffsetMap.Add(62, "B8");

            PianoKey.NoteOffsetMap.Add(63, "C9");
            PianoKey.NoteOffsetMap.Add(64, "C#9");
            PianoKey.NoteOffsetMap.Add(65, "D9");
            PianoKey.NoteOffsetMap.Add(66, "D#9");
            PianoKey.NoteOffsetMap.Add(67, "E9");
            PianoKey.NoteOffsetMap.Add(68, "F9");
            PianoKey.NoteOffsetMap.Add(69, "F#9");
            PianoKey.NoteOffsetMap.Add(70, "G9");
            PianoKey.NoteOffsetMap.Add(71, "G#9");
            PianoKey.NoteOffsetMap.Add(72, "A9");
            PianoKey.NoteOffsetMap.Add(73, "A#9");
            PianoKey.NoteOffsetMap.Add(74, "B9");
        }

        public float CalculateNoteFrequency(float a4hertz, int distance)
        {
            return (float)System.Math.Pow(2.0f,((float)distance / 12.0f)) * a4hertz;
        }

        public float CalculateMilliseconds(float notevalue, float bpm)
        {
            return (60f / bpm) * notevalue * 1000;
        }

        public float GetCurrentNoteValue()
        {
            /*
            if (radioButton1.Checked)
                return 1.0f / 128.0f;
            else if (radioButton2.Checked)
                return 1.0f / 64.0f;
            else if (radioButton3.Checked)
                return 1.0f / 32.0f;
            else if (radioButton4.Checked)
                return 1.0f / 16.0f;
            else if (radioButton5.Checked)
                return 1.0f / 8.0f;
            else if (radioButton6.Checked)
                return 1.0f / 4.0f;
            else if (radioButton7.Checked)
                return 1.0f / 2.0f;
            else if (radioButton8.Checked)
                return 1.0f;
            else
                return 1.0f / 4.0f;
            */

            return (float)(4*(numLenTop.Value / numLenBottom.Value));
        }
                
        private void button91_Click(object sender, EventArgs e)
        {
            Button b = (Button)sender;

            int distance = Convert.ToInt32(b.Tag);

            unsavedChanges = true;
            UpdateCaption();

            lstNotes.Items.Add(new Note() { notevalue = GetCurrentNoteValue(), distance = distance, sleep = false });
        
            if (chkAutoRest.Checked)
                lstNotes.Items.Add(new Note() { distance = 0, notevalue = GetCurrentNoteValue(), sleep = true });
        }

        private void pianoKeyboard1_PianoKeyPressed(object sender, PianoKeyPressedEventArgs e)
        {
            lstNotes.Items.Add(new Note() { notevalue = GetCurrentNoteValue(), distance = e.NoteOffset, sleep = false });

            unsavedChanges = true;
            UpdateCaption();

            if (chkAutoRest.Checked)
                lstNotes.Items.Add(new Note() { distance = 0, notevalue = GetCurrentNoteValue(), sleep = true });
        }

        private void Main_Load(object sender, EventArgs e)
        {
            ds = new sndDirectSound(this);
            waveGen = new WaveGen(ds);
            PopulateWaveforms();
            NewData();
        }

        private void PopulateWaveforms() {
            waveforms.Add(new WaveformType("Sine", WaveGen.Waveform.Sine));
            waveforms.Add(new WaveformType("Sine^2", WaveGen.Waveform.SinePow2));
            waveforms.Add(new WaveformType("Sine^4", WaveGen.Waveform.SinePow4));
            waveforms.Add(new WaveformType("Square (5% duty)", WaveGen.Waveform.Square5));
            waveforms.Add(new WaveformType("Square (10% duty)", WaveGen.Waveform.Square10));
            waveforms.Add(new WaveformType("Square (15% duty)", WaveGen.Waveform.Square15));
            waveforms.Add(new WaveformType("Square (20% duty)", WaveGen.Waveform.Square20));
            waveforms.Add(new WaveformType("Square (25% duty)", WaveGen.Waveform.Square25));
            waveforms.Add(new WaveformType("Square (30% duty)", WaveGen.Waveform.Square30));
            waveforms.Add(new WaveformType("Square (35% duty)", WaveGen.Waveform.Square35));
            waveforms.Add(new WaveformType("Square (40% duty)", WaveGen.Waveform.Square40));
            waveforms.Add(new WaveformType("Square (45% duty)", WaveGen.Waveform.Square45));
            waveforms.Add(new WaveformType("Square (50% duty)", WaveGen.Waveform.Square50));
            waveforms.Add(new WaveformType("Triangle", WaveGen.Waveform.Triangle));
            waveforms.Add(new WaveformType("Sawtooth", WaveGen.Waveform.Sawtooth));
            waveforms.Add(new WaveformType("Noise (white)", WaveGen.Waveform.WhiteNoise));

            foreach(WaveformType wave in waveforms)
                cboWaveform.Items.Add(wave);

            cboWaveform.SelectedIndex = 0;
        }

        private void button_AddRest_Click(object sender, EventArgs e)
        {
            lstNotes.Items.Add(new Note() { distance = 0, notevalue = GetCurrentNoteValue(), sleep = true });
            unsavedChanges = true;
            UpdateCaption();
        }

        private void button_RemoveNote_Click(object sender, EventArgs e)
        {
            RemoveNote();
            unsavedChanges = true;
            UpdateCaption();
        }

        private void RemoveNote() {
            if(lstNotes.Items.Count == 0) return;

            if(lstNotes.SelectedItems.Count >= 1)
                while(lstNotes.SelectedItems.Count != 0)
                    lstNotes.Items.Remove(lstNotes.SelectedItems[0]);
            else  // remove the last entered note
                lstNotes.Items.RemoveAt(lstNotes.Items.Count - 1);

            unsavedChanges = true;
            UpdateCaption();
        }

        private void Main_KeyPress(object sender, KeyPressEventArgs e)
        {
           
        }

        private void Main_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                RemoveNote();
            }
        }

        private void lstNotes_KeyDown(object sender, KeyEventArgs e) {
            if(e.KeyCode == Keys.Delete) RemoveNote();
        }

        private void NewData() {
            initializing = true;
            chanCount = currentChannel = 0;
            channels.Clear();
            AddChannel();
            lstNotes.Items.Clear();
            txtName.Text = "Untitled Song";
            numA4.Value = 440;
            numAttack.Value = 0.005M;
            numDecay.Value = 0.010M;
            numChanVol.Value = 100;
            numBPM.Value = 120;
            numLenTop.Value = 1;
            numLenBottom.Value = 4;
            chkAutoRest.Checked = false;
            unsavedChanges = false;
            activePath = "";
            activeFile = "Untitled.beep";
            UpdateCaption();
            initializing = false;
        }

        private void AddChannel() {
            newChanInit = true;
            Channel newChan = new Channel();
            newChan.waveform = WaveGen.Waveform.Sine;
            newChan.patterns = new ArrayList();
            newChan.order = new ArrayList();
            newChan.volume = 100;
            newChan.attack = 0.005M;
            newChan.decay = 0.010M;
            newChan.name = String.Format("Channel {0}", ++chanCount);
            channels.Add(newChan);
            RefreshChannels();
            cboChannel.SelectedIndex = cboChannel.Items.Count - 1;
            currentPattern = 0;
            AddPattern();
            lstPatterns.SelectedIndex = 0;
            newChanInit = false;
            lstOrder.Items.Add(Convert.ToInt32(0));

            unsavedChanges = true;
            UpdateCaption();
        }

        private void AddPattern() {
            Channel chan = (Channel)channels[currentChannel];
            Pattern pattern = new Pattern();
            int newID = chan.GetNextFreeID();
            pattern.id = newID;
            pattern.notes = new ArrayList();
            chan.patterns.Add(pattern);
            chan.SortPatterns();
            channels[currentChannel] = chan;  // save to update pattern sort
            RefreshPatterns();

            foreach(int index in lstPatterns.Items) {
                if(index == newID) {
                    lstPatterns.SelectedItem = index;
                    break;
                }
            }
        }

        private void AddPattern(int newID) {
            Channel chan = (Channel)channels[currentChannel];

            foreach(Pattern pat in chan.patterns)
                if(pat.id == newID) return;

            Pattern pattern = new Pattern();
            pattern.id = newID;
            pattern.notes = new ArrayList();
            chan.patterns.Add(pattern);
            chan.SortPatterns();
            channels[currentChannel] = chan;  // save to update pattern sort
            RefreshPatterns();

            foreach(int index in lstPatterns.Items) {
                if(index == newID) {
                    lstPatterns.SelectedItem = index;
                    break;
                }
            }
        }

        private void RefreshChannels() {
            cboChannel.Items.Clear();
            foreach(Channel chan in channels)
                cboChannel.Items.Add(chan.name);

            btnRemoveChannel.Enabled = !(cboChannel.Items.Count == 1);
        }

        private void RefreshPatterns() {
            lstPatterns.Items.Clear();
            Channel chan = (Channel)channels[currentChannel];
            foreach(Pattern pat in chan.patterns)
                lstPatterns.Items.Add(pat.id);

            btnPatternDel.Enabled = !(lstPatterns.Items.Count == 1);
        }

        private void RefreshCurrentPattern() {
            lstNotes.Items.Clear();
            Pattern pat = (Pattern)((Channel)channels[currentChannel]).patterns[currentPattern];
            foreach(Note note in pat.notes)
                lstNotes.Items.Add(note);
        }

        private void btnAddChannel_Click(object sender, EventArgs e) {
            AddChannel();

            unsavedChanges = true;
            UpdateCaption();
        }

        private void btnRemoveChannel_Click(object sender, EventArgs e) {
            deleting = true;
            int selectedIndex = cboChannel.SelectedIndex;
            if(cboChannel.Items.Count <= 1) return;
            channels.RemoveAt(cboChannel.SelectedIndex);
            RefreshChannels();
            if(selectedIndex == cboChannel.Items.Count) selectedIndex--;
            cboChannel.SelectedIndex = selectedIndex;

            unsavedChanges = true;
            UpdateCaption();
        }

        private void btnRename_Click(object sender, EventArgs e) {
            Channel chan = (Channel)channels[cboChannel.SelectedIndex];
            dlgRename renameDialog = new dlgRename();
            renameDialog.NewName = chan.name;
            if(renameDialog.ShowDialog(this) == DialogResult.OK) {
                chan.name = renameDialog.NewName;
                channels[cboChannel.SelectedIndex] = chan;
                RefreshChannels();
                cboChannel.SelectedIndex = currentChannel;
            }

            unsavedChanges = true;
            UpdateCaption();
        }

        private void SaveChannel() {
            // don't try to save if we're deleting the channel
            if(deleting) {
                deleting = false;
                return;
            }

            // save pattern before saving channel
            SavePattern();

            Channel chan = new Channel();
            Channel oldChan = (Channel)channels[currentChannel];
            chan.patterns = new ArrayList();
            chan.order = new ArrayList();

            foreach(Pattern pat in oldChan.patterns)
                chan.patterns.Add(pat);

            foreach(int patID in lstOrder.Items)
                chan.order.Add(patID);

            chan.name = ((Channel)channels[currentChannel]).name;
            chan.volume = (byte)numChanVol.Value;
            chan.attack = numAttack.Value;
            chan.decay = numDecay.Value;
            chan.waveform = ((WaveformType)cboWaveform.SelectedItem).Waveform;
            channels[currentChannel] = chan;
        }

        private void SavePattern() {
            // don't try to save if we're deleting the pattern
            if(deleting) {
                deleting = false;
                return;
            }

            Channel chan = (Channel)channels[currentChannel];
            Pattern pat = new Pattern();
            pat.id = Convert.ToInt32(lstPatterns.Items[currentPattern]);
            pat.notes = new ArrayList();
            foreach(Note note in lstNotes.Items)
                pat.notes.Add(note);

            chan.patterns[currentPattern] = pat;
            channels[currentChannel] = chan;
        }

        private void cboChannel_SelectedIndexChanged(object sender, EventArgs e) {
            // save current channel
            if(!initializing) SaveChannel();

            // clear lists
            lstNotes.Items.Clear();
            lstPatterns.Items.Clear();

            // now change to the desired one
            currentChannel = cboChannel.SelectedIndex;
            Channel chan = (Channel)channels[currentChannel];
            numChanVol.Value = chan.volume;
            numAttack.Value = chan.attack;
            numDecay.Value = chan.decay;

            foreach(WaveformType wave in cboWaveform.Items) {
                lstOrder.Items.Clear();
                if(chan.waveform == wave.Waveform) {
                    cboWaveform.SelectedItem = wave;
                    break;
                }
            }

            currentPattern = 0;
            RefreshPatterns();

            if(!initializing && !newChanInit) {
                foreach(int patID in chan.order)
                    lstOrder.Items.Add(patID);
                LoadCurrentPattern();
            }
        }

        private void lstPatterns_SelectedIndexChanged(object sender, EventArgs e) {
            // save current pattern
            if(!initializing) SavePattern();

            // load new pattern
            Channel chan = (Channel)channels[currentChannel];
            currentPattern = lstPatterns.SelectedIndex;
            Pattern pat = (Pattern)chan.patterns[currentPattern];
            
            // load notes
            lstNotes.Items.Clear();
            foreach(Note note in pat.notes)
                lstNotes.Items.Add(note);
        }

        private void LoadCurrentPattern() {
            // load pattern
            Channel chan = (Channel)channels[currentChannel];
            Pattern pat = (Pattern)chan.patterns[currentPattern];

            // load notes
            lstNotes.Items.Clear();
            foreach(Note note in pat.notes)
                lstNotes.Items.Add(note);

            lstPatterns.SelectedIndex = currentPattern;
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e) {
            DialogResult dlg;
            dlg = UnsavedChangesPrompt();
            if(dlg == DialogResult.Cancel) return;
            if(dlg == DialogResult.Yes) Save(false);
            NewData();
            UpdateCaption();
        }

        private void LoadV1(StreamReader sr, string title) {
            // load globals
            txtName.Text = title;
            numA4.Value = Convert.ToDecimal(sr.ReadLine());
            numBPM.Value = Convert.ToDecimal(sr.ReadLine());

            int numChannels = Convert.ToInt32(sr.ReadLine());

            // load channels
            for(int i=0; i < numChannels; i++) {
                // read channel header
                Channel chan = new Channel();
                chan.name = sr.ReadLine();
                chan.volume = Convert.ToByte(sr.ReadLine());
                chan.attack = Convert.ToDecimal(sr.ReadLine());
                chan.decay = Convert.ToDecimal(sr.ReadLine());
                chan.patterns = new ArrayList();
                chan.order = new ArrayList();

                // read channel waveform
                string inWaveform = sr.ReadLine();

                // the name of the square wave has changed -- if it's "square," pick the right one ("Square50")
                if(inWaveform == "Square")
                    chan.waveform = WaveGen.Waveform.Square50;
                else {
                    // otherwise, pick it as we normally would
                    foreach(WaveformType wave in waveforms) {
                        if(inWaveform == wave.Name) {
                            chan.waveform = wave.Waveform;
                            break;
                        }
                    }
                }

                int numNotes = Convert.ToInt32(sr.ReadLine());

                Pattern pat = new Pattern();
                pat.id = 0;
                pat.notes = new ArrayList();

                // read notes
                for(int j=0; j < numNotes; j++) {
                    Note note = new Note();
                    string inNote = sr.ReadLine();
                    if(inNote == "r") note.sleep = true;
                    else note.distance = Convert.ToInt32(inNote);
                    note.notevalue = Convert.ToSingle(sr.ReadLine());
                    pat.notes.Add(note);
                }

                chan.patterns.Add(pat);
                chan.order.Add(0);

                channels.Add(chan);
            }

            sr.Close();
        }

        private void LoadV2(StreamReader sr) {
            // load globals
            txtName.Text = sr.ReadLine();
            numA4.Value = Convert.ToDecimal(sr.ReadLine());
            numBPM.Value = Convert.ToDecimal(sr.ReadLine());

            int numChannels = Convert.ToInt32(sr.ReadLine());

            // load channels
            for(int i=0; i < numChannels; i++) {
                // read channel header
                Channel chan = new Channel();
                chan.name = sr.ReadLine();
                chan.volume = Convert.ToByte(sr.ReadLine());
                chan.attack = Convert.ToDecimal(sr.ReadLine());
                chan.decay = Convert.ToDecimal(sr.ReadLine());
                chan.patterns = new ArrayList();
                chan.order = new ArrayList();

                // read channel waveform
                string inWaveform = sr.ReadLine();
                
                // the name of the square wave has changed -- if it's "square," pick the right one ("Square50")
                if(inWaveform == "Square")
                    chan.waveform = WaveGen.Waveform.Square50;
                else {
                    // otherwise, pick it as we normally would
                    foreach(WaveformType wave in waveforms) {
                        if(inWaveform == wave.Name) {
                            chan.waveform = wave.Waveform;
                            break;
                        }
                    }
                }

                int numPats = Convert.ToInt32(sr.ReadLine());

                // read patterns
                for(int j=0; j < numPats; j++) {
                    Pattern pat = new Pattern();

                    pat.id = Convert.ToInt32(sr.ReadLine());
                    pat.notes = new ArrayList();
                    int numNotes = Convert.ToInt32(sr.ReadLine());

                    // read notes
                    for(int k=0; k < numNotes; k++) {
                        Note note = new Note();
                        string inNote = sr.ReadLine();
                        if(inNote == "r") note.sleep = true;
                        else note.distance = Convert.ToInt32(inNote);
                        note.notevalue = Convert.ToSingle(sr.ReadLine());
                        pat.notes.Add(note);
                    }

                    chan.patterns.Add(pat);
                }

                // read order
                int numOrderEnts = Convert.ToInt32(sr.ReadLine());

                for(int j=0; j < numOrderEnts; j++)
                    chan.order.Add(Convert.ToInt32(sr.ReadLine()));

                channels.Add(chan);
            }

            sr.Close();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e) {
            string version;
            StreamReader sr;

            DialogResult dlg = dlgOpen.ShowDialog();
            if(dlg == DialogResult.Cancel) return;

            dlg = UnsavedChangesPrompt();
            if(dlg == DialogResult.Cancel) return;
            if(dlg == DialogResult.Yes) Save(false);

            // open file
            try {
                sr = new StreamReader(dlgOpen.OpenFile());
            } catch {
                MessageBox.Show("Error opening file for reading!", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            // throw everything out :o
            deleting = true;
            cboChannel.Items.Clear();
            channels.Clear();
            chanCount = 0;
            currentChannel = 0;
            lstNotes.Items.Clear();

            // read file version
            version = sr.ReadLine();

            if(version != "BMv2") {
                // for V1 files, the "version" field would be the song title
                LoadV1(sr, version);
            } else {
                LoadV2(sr);
            }

            // refresh channel list and jump to the first channel
            RefreshChannels();
            deleting = true;
            cboChannel.SelectedIndex = 0;

            // update active file and caption
            activePath = dlgOpen.FileName;
            activeFile = Path.GetFileName(dlgOpen.FileName);
            unsavedChanges = false;
            UpdateCaption();
        }

        private bool Save(bool saveAs) {
            StreamWriter sw;

            // save the channel data to memory first
            SaveChannel();

            if(saveAs == true || activePath == "") {
                DialogResult dlg = dlgSave.ShowDialog();
                if(dlg == DialogResult.Cancel) return false;
                sw = new StreamWriter(dlgSave.OpenFile());
                activePath = dlgSave.FileName;
                activeFile = Path.GetFileName(activePath);
            } else
                sw = new StreamWriter(activePath);

            // write file version
            sw.WriteLine("BMv2");

            // write global song data
            sw.WriteLine(txtName.Text);
            sw.WriteLine(numA4.Value);
            sw.WriteLine(numBPM.Value);
            sw.WriteLine(channels.Count);

            // write channel data
            foreach(Channel chan in channels) {
                // write channel header
                sw.WriteLine(chan.name);
                sw.WriteLine(chan.volume);
                sw.WriteLine(chan.attack);
                sw.WriteLine(chan.decay);

                // determine waveform name
                foreach(WaveformType wave in waveforms) {
                    if(chan.waveform == wave.Waveform) {
                        sw.WriteLine(wave.Name);
                        break;
                    }
                }

                // write number of patterns
                sw.WriteLine(chan.patterns.Count);

                foreach(Pattern pat in chan.patterns) {
                    // write pattern header and note count
                    sw.WriteLine(pat.id);
                    sw.WriteLine(pat.notes.Count);

                    foreach(Note note in pat.notes) {
                        sw.WriteLine(note.sleep ? "r" : note.distance.ToString());
                        sw.WriteLine(note.notevalue);
                    }
                }

                // write order count and entries
                sw.WriteLine(chan.order.Count);

                foreach(int patID in chan.order)
                    sw.WriteLine(patID);
            }

            // close the file
            sw.Close();

            unsavedChanges = false;
            UpdateCaption();

            return true;
        }

        private void ExportWAV() {
            Stream fs;

            // save the channel data to memory first
            SaveChannel();

            DialogResult dlg = dlgSaveWAV.ShowDialog();
            if(dlg == DialogResult.Cancel) return;
            fs = dlgSaveWAV.OpenFile();
            waveGen.Render(channels, numBPM.Value, numA4.Value);
            waveGen.ExportWAV(fs);
            fs.Close();
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e) {
            Save(false);
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e) {
            Save(true);
        }

        private void UpdateCaption() {
            string caption = "BeepMap - [" + activeFile + "]";
            if(unsavedChanges) caption += " (unsaved)";
            this.Text = caption;
        }

        private DialogResult UnsavedChangesPrompt() {
            if(unsavedChanges)
                return MessageBox.Show("Do you want to save the changes that you made to the current song?", "BeepMap", MessageBoxButtons.YesNoCancel);
            else
                return DialogResult.No;
        }

        private void exportWAVToolStripMenuItem_Click(object sender, EventArgs e) {
            ExportWAV();
        }

        private void renderPatternToolStripMenuItem_Click(object sender, EventArgs e) {
            SaveChannel();
            Pattern tempPat = new Pattern();
            tempPat.id = 0;
            tempPat.notes = new ArrayList();
            foreach(Note note in lstNotes.Items)
                tempPat.notes.Add(note);
            waveGen.RenderPattern((Channel)channels[currentChannel], tempPat, numBPM.Value, numA4.Value);
        }

        private void renderSongToolStripMenuItem_Click(object sender, EventArgs e) {
            SaveChannel();
            waveGen.Render(channels, numBPM.Value, numA4.Value);
        }

        private void renderChannelStripMenuItem3_Click(object sender, EventArgs e) {
            SavePattern();
            SaveChannel();
            ArrayList tempChanList = new ArrayList();
            tempChanList.Add((Channel)channels[currentChannel]);
            waveGen.Render(tempChanList, numBPM.Value, numA4.Value);
        }

        private void playToolStripMenuItem_Click(object sender, EventArgs e) {
            waveGen.Play();
        }

        private void stopToolStripMenuItem_Click(object sender, EventArgs e) {
            waveGen.Stop();
        }

        private void numChanVol_ValueChanged(object sender, EventArgs e) {
            unsavedChanges = true;
            UpdateCaption();
        }

        private void cboWaveform_SelectedIndexChanged(object sender, EventArgs e) {
            unsavedChanges = true;
            UpdateCaption();
        }

        private void numAttack_ValueChanged(object sender, EventArgs e) {
            unsavedChanges = true;
            UpdateCaption();
        }

        private void numDecay_ValueChanged(object sender, EventArgs e) {
            unsavedChanges = true;
            UpdateCaption();
        }

        private void btnPatternAdd_Click(object sender, EventArgs e) {
            AddPattern();
        }

        private void btnPatternDel_Click(object sender, EventArgs e) {
            Channel chan = (Channel)channels[currentChannel];
            deleting = true;
            int selectedIndex = lstPatterns.SelectedIndex;
            int deletedID = (Int32)lstPatterns.SelectedItem;
            if(lstPatterns.Items.Count <= 1) return;
            chan.patterns.RemoveAt(lstPatterns.SelectedIndex);
            RefreshPatterns();
            if(selectedIndex == lstPatterns.Items.Count) selectedIndex--;
            lstPatterns.SelectedIndex = selectedIndex;

            // update order list
            for(int i=lstOrder.Items.Count-1; i > 0; i--) {
                if((Int32)lstOrder.Items[i] == deletedID)
                    lstOrder.Items.RemoveAt(i);
            }

            unsavedChanges = true;
            UpdateCaption();
        }

        private void lstPatterns_DoubleClick(object sender, EventArgs e) {
            if(lstPatterns.SelectedIndex >= 0)
                AddToOrder(lstPatterns.SelectedIndex);
        }

        private void AddToOrder(int index) {
            lstOrder.Items.Add(lstPatterns.Items[index]);
            TestButtonsOrder();
            lstOrder.SelectedIndex = lstOrder.Items.Count - 1;
        }
        
        private void btnOrderAdd_Click(object sender, EventArgs e) {
            AddToOrder(lstPatterns.SelectedIndex);
        }

        private void btnOrderRemove_Click(object sender, EventArgs e) {
            int selectedIndex = lstOrder.SelectedIndex;
            lstOrder.Items.Remove(lstOrder.SelectedItem);
            if(selectedIndex == lstOrder.Items.Count)
                lstOrder.SelectedIndex = lstOrder.Items.Count - 1;
            else
                lstOrder.SelectedIndex = selectedIndex;
            TestButtonsOrder();
        }

        private void btnOrderUp_Click(object sender, EventArgs e) {
            int oldIndex = lstOrder.SelectedIndex;
            int newIndex = lstOrder.SelectedIndex - 1;
            int oldValue = (Int32)lstOrder.Items[newIndex];
            lstOrder.Items[newIndex] = (Int32)lstOrder.SelectedItem;
            lstOrder.Items[oldIndex] = oldValue;
            lstOrder.SelectedIndex = newIndex;
        }

        private void btnOrderDown_Click(object sender, EventArgs e) {
            int oldIndex = lstOrder.SelectedIndex;
            int newIndex = lstOrder.SelectedIndex + 1;
            int oldValue = (Int32)lstOrder.Items[newIndex];
            lstOrder.Items[newIndex] = (Int32)lstOrder.SelectedItem;
            lstOrder.Items[oldIndex] = oldValue;
            lstOrder.SelectedIndex = newIndex;
        }

        private void TestButtonsOrder() {
            if(lstOrder.SelectedIndex == 0)
                btnOrderUp.Enabled = false;
            else
                btnOrderUp.Enabled = true;

            if(lstOrder.SelectedIndex == lstOrder.Items.Count - 1)
                btnOrderDown.Enabled = false;
            else
                btnOrderDown.Enabled = true;
        }

        private void lstOrder_SelectedIndexChanged(object sender, EventArgs e) {
            TestButtonsOrder();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e) {
            this.Close();
        }

        private void Main_FormClosing(object sender, FormClosingEventArgs e) {
            DialogResult dlg = UnsavedChangesPrompt();

            if(dlg == DialogResult.Cancel)
                e.Cancel = true;
            else if(dlg == DialogResult.Yes)
                if(!Save(false))
                    e.Cancel = true;
        }

        private void Transpose(int distance) {
            Pattern pat = (Pattern)((Channel)channels[currentChannel]).patterns[currentPattern];  // say that three times fast
            Pattern newPat = new Pattern();
            
            newPat.id = pat.id;
            newPat.notes = new ArrayList();

            foreach(Note note in pat.notes) {
                Note newNote = note;
                newNote.distance += distance;

                // make sure that no notes go above/below the limits
                if(newNote.distance > 50 || newNote.distance < -45)
                    return;

                newPat.notes.Add(newNote);
            }

            // save pattern to channel and refresh note list
            ((Channel)channels[currentChannel]).patterns[currentPattern] = newPat;
            RefreshCurrentPattern();
        }

        private void octaveDownToolStripMenuItem_Click(object sender, EventArgs e) {
            Transpose(-12);
        }

        private void octaveUpToolStripMenuItem_Click(object sender, EventArgs e) {
            Transpose(12);
        }

        private void semitoneDownToolStripMenuItem_Click(object sender, EventArgs e) {
            Transpose(-1);
        }

        private void semitoneUpToolStripMenuItem_Click(object sender, EventArgs e) {
            Transpose(1);
        }
    }

    internal class WaveformType {
        private string name;
        private WaveGen.Waveform waveform;

        public WaveGen.Waveform Waveform {
            get { return waveform; }
            set { waveform = value; }
        }

        public string Name {
            get { return name; }
            set { name = value; }
        }

        public WaveformType() { }

        public WaveformType(string inName, WaveGen.Waveform inWaveform) {
            name = inName;
            waveform = inWaveform;
        }

        public override string ToString() {
            return name;
        }
    }
}
