﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;

namespace BeepMap
{
    public class PianoKey : PictureBox
    {
        /* -- Note <-> Offset Map -- */
        public static Hashtable NoteOffsetMap = new Hashtable();
                        
        /* -- Constants -- */
        public Image KeyL = Properties.Resources.Key_L;
        public Image KeyL_Down = Properties.Resources.Key_L_Down;

        public Image KeyFlat = Properties.Resources.Key_Rectangle;
        public Image KeyFlat_Down = Properties.Resources.Key_Rectangle_Down;

        public Image KeyT = Properties.Resources.Key_T;
        public Image KeyT_Down = Properties.Resources.Key_T_Down;

        public Image KeyBackwardsL = Properties.Resources.Key_L_Backwards;
        public Image KeyBackwardsL_Down = Properties.Resources.Key_L_Backwards_Down;
                        
        /* -- Fields -- */
        private bool keyDown;
        private PianoKeyShape shape;

        /* -- Properties -- */
        public Image ImageDown { get; set; }

        public bool PianoKeyDown
        {
            get { return keyDown; }
            set { keyDown = value; }
        }

        public int NoteOffset { get; set; }

        public PianoKeyShape Shape
        {
            get { return shape; }
            set
            {
                switch (value)
                {
                    case PianoKeyShape.L:
                        this.Image = KeyL;
                        this.ImageDown = KeyL_Down;
                        break;

                    case PianoKeyShape.Flat:
                        this.Image = KeyFlat;
                        this.ImageDown = KeyFlat_Down;
                        break;

                    case PianoKeyShape.T:
                        this.Image = KeyT;
                        this.ImageDown = KeyT_Down;
                        break;

                    case PianoKeyShape.BackwardsL:
                        this.Image = KeyBackwardsL;
                        this.ImageDown = KeyBackwardsL_Down;
                        break;
                }

                shape = value;
            }
        }
        
        // Designer-required
        private System.ComponentModel.Container components = null;
                
        /* -- Constructor -- */
        public PianoKey()
        {
            // This call is required by the Windows.Forms Form Designer.
            InitializeComponent();
                        
            //
            // Initialize properties.
            //
            this.Shape = PianoKeyShape.L;
            this.PianoKeyDown = false;            
        }
                
        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (components != null)
                    components.Dispose();
            }
            base.Dispose(disposing);
        }

        #region Component Designer generated code
        /// <summary>
        /// Required method for Designer support - do not modify 
        /// the contents of this method with the code editor.
        /// </summary>
        private void InitializeComponent()
        {
            ((System.ComponentModel.ISupportInitialize)(this)).BeginInit();
            this.SuspendLayout();
            // 
            // PianoKey
            // 
            this.MouseClick += new System.Windows.Forms.MouseEventHandler(this.PianoKey_MouseClick);
            this.MouseDown += new System.Windows.Forms.MouseEventHandler(this.PianoKey_MouseDown);
            this.MouseUp += new System.Windows.Forms.MouseEventHandler(this.PianoKey_MouseUp);
            ((System.ComponentModel.ISupportInitialize)(this)).EndInit();
            this.ResumeLayout(false);

        }
        #endregion                

        /* -- Methods & Events -- */
        private void UpdateKeyImage()
        {
            if (this.keyDown)
            {
                switch (this.shape)
                {
                    case PianoKeyShape.L:
                        this.ImageDown = KeyL;
                        this.Image = KeyL_Down;
                        break;

                    case PianoKeyShape.Flat:
                        this.ImageDown = KeyFlat;
                        this.Image = KeyFlat_Down;
                        break;

                    case PianoKeyShape.T:
                        this.ImageDown = KeyT;
                        this.Image = KeyT_Down;
                        break;

                    case PianoKeyShape.BackwardsL:
                        this.ImageDown = KeyBackwardsL;
                        this.Image = KeyBackwardsL_Down;
                        break;
                }
            }
            else
            {
                switch (this.shape)
                {
                    case PianoKeyShape.L:
                        this.Image = KeyL;
                        this.ImageDown = KeyL_Down;
                        break;

                    case PianoKeyShape.Flat:
                        this.Image = KeyFlat;
                        this.ImageDown = KeyFlat_Down;
                        break;

                    case PianoKeyShape.T:
                        this.Image = KeyT;
                        this.ImageDown = KeyT_Down;
                        break;

                    case PianoKeyShape.BackwardsL:
                        this.Image = KeyBackwardsL;
                        this.ImageDown = KeyBackwardsL_Down;
                        break;
                }
            }
        }

        private void PianoKey_MouseClick(object sender, MouseEventArgs e)
        {
            this.PianoKeyDown = true;
            UpdateKeyImage();
        }

        private void PianoKey_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                this.PianoKeyDown = true;
                UpdateKeyImage();
            }

            //base.OnMouseDown(e);
        }

        private void PianoKey_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                this.PianoKeyDown = false;
                UpdateKeyImage();
            }

            //base.OnMouseUp(e);
        }
    }

    public enum PianoKeyShape
    {
        L,
        Flat,
        T,
        BackwardsL
    }
}
