﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;

namespace BeepMap
{
    public partial class PianoKeyboard : UserControl
    {
        public event PianoKeyPressedHandler PianoKeyPressed;

        private bool[] keyStates;

        public int Octave
        {
            get
            {
                return octaveScrollBar.Value;
            }
            set
            {
                octaveScrollBar.Value = value;
            }
        }

        public PianoKeyboard()
        {
            InitializeComponent();

            keyStates = new bool[29];
        }

        protected virtual void OnPianoKeyPressed(PianoKeyPressedEventArgs e)
        {
            PianoKeyPressed(this, e);
        }

        private void PianoKeyboard_Load(object sender, EventArgs e)
        {
            octaveLabel.Text = String.Format("Octave: {0}", octaveScrollBar.Value);
        }
                
        private int CalculateNoteOffset(int offset)
        {
            return 12 * (octaveScrollBar.Value - 4) + offset;
        }

        private void PianoKey_MouseUp(object sender, MouseEventArgs e)
        {
            PianoKey key = (PianoKey)sender;

            OnPianoKeyPressed(new PianoKeyPressedEventArgs() { NoteOffset = CalculateNoteOffset(key.NoteOffset) });
        }

        private void octaveScrollBar_Scroll(object sender, ScrollEventArgs e)
        {
            octaveLabel.Text = String.Format("Octave: {0}", octaveScrollBar.Value);
        }

        private void PianoKey_MouseHover(object sender, EventArgs e)
        {
            noteLabel.Text = String.Format("Note: {0}", PianoKey.NoteOffsetMap[CalculateNoteOffset(((PianoKey)sender).NoteOffset)]);
        }               
    }

    public class PianoKeyPressedEventArgs : EventArgs
    {
        public int NoteOffset { get; set; }
    }

    public delegate void PianoKeyPressedHandler(object sender, PianoKeyPressedEventArgs e);
}
