﻿/*
 * Copyright (c) 2010 Ian "Spectere" Burgmyer
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event will the authors be held liable for any damages
 * arising from the use of this software.
 * 
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 * 
 * 1. The origin of this software must not be misrepresented; you must not
 * claim that you wrote the original software. If you use this software
 * in a product, an acknowledgment in the product documentation would be
 * appreciated but is not required.
 * 
 * 2. Altered source versions must be plainly marked as such, and must not be
 * misrepresented as being the original software.
 * 
 * 3. This notice may not be removed or altered from any source
 * distribution.
 */

using Synthia.Plugin;
using Synthia.Plugin.Generator;
using System;

namespace Synthia.Plugin.Generator.SynthiaSine {
    [Plugin(PluginType.Generator)]
    public class SynthiaSine : IGenerator {
        public string Name { get { return "sine"; } }
        public string FullName { get { return "SynthiaSine"; } }
        public string Author { get { return "Spectere"; } }
        public string Version { get { return "1.0"; } }
        public string URL { get { return "http://www.spectere.net/"; } }

        public void Generate(ref Int64[] left, ref Int64[] right, UInt32 sampleRate, double time, UInt16 bpm, UInt32 startBeat, UInt32 endBeat, UInt16 startFreq, UInt16 endFreq, double startVolume, double endVolume, Int16 startPan, Int16 endPan) {
            double frequency = startFreq, panning = startPan, volume = startVolume, panL = 1.0, panR = 1.0;
            UInt64 samples = (UInt64)((time / 1000) * (double)sampleRate);

            Int64[] samplesL = new Int64[samples];
            Int64[] samplesR = new Int64[samples];

            for(UInt64 i = 0; i < samples; i++) {
                frequency = (double)startFreq - ((double)startFreq - (double)endFreq) * ((double)i / (((double)samples*2)-1));
                volume = (double)startVolume - ((double)startVolume - (double)endVolume) * ((double)i / ((double)samples-1));
                panning = (double)startPan - ((double)startPan - (double)endPan) * ((double)i / ((double)samples-1));
                if(panning > 0) {           // pan toward the right
                    panL = 1.0 - ((double)Math.Abs(panning) / 100);
                } else if(panning < 0) {    // pan audio toward the left
                    panR = 1.0 - ((double)Math.Abs(panning) / 100);
                }

                double amplitudeL = (double)Int64.MaxValue * volume * panL;
                double amplitudeR = (double)Int64.MaxValue * volume * panR;
                double t = (Math.PI * 2 * frequency) / sampleRate;

                samplesL[i] = (Int64)Math.Round(amplitudeL * Math.Sin(i * t), 0);
                samplesR[i] = (Int64)Math.Round(amplitudeR * Math.Sin(i * t), 0);
            }

            left = samplesL;
            right = samplesR;
        }
    }
}
