﻿using System;

namespace BeepMap {
    partial class WaveGen {
        private double[] WaveSine(double frequency, int numSamples, double volume, double pow) {
            double t = (Math.PI * 2 * frequency) / sampleRate;
            double[] samples = new double[numSamples];
            for(int i = 0; i < numSamples; i++)
                samples[i] = volume * Math.Pow(Math.Sin(i * t), pow);

            return samples;
        }

        private double[] WaveSquare(double frequency, int numSamples, double volume, double duty) {
            double[] samples = new double[numSamples];
            double cycle = sampleRate / frequency;
            double cycleOn = cycle * duty;
            double cycleOff = cycle - cycleOn;

            for(int i = 0; i < numSamples; i++) {
                samples[i] = ((i % cycle) <= cycleOn) ? volume : -volume;
            }

            return samples;
        }

        private double[] WaveSawtooth(double frequency, int numSamples, double volume) {
            double wavelength = sampleRate / frequency;
            double step = (volume * 2) / wavelength;
            double sample = -volume;
            double[] samples = new double[numSamples];
            int samplesOut = 0;

            while(samplesOut < numSamples) {
                sample = -volume;

                for(int i=0; i < wavelength && samplesOut < numSamples; i++) {
                    sample += step;
                    if(sample > 1) sample = 1;
                    if(sample < -1) sample = -1;
                    samples[samplesOut] = sample;
                    samplesOut++;
                }
            }

            return samples;
        }

        private double[] WaveTriangle(double frequency, int numSamples, double volume) {
            double wavelength = sampleRate / frequency;
            double step = (volume * 2) / wavelength;
            double sample = -volume;
            double[] samples = new double[numSamples];

            for(int i=0; i < numSamples - 1; i++) {
                if(Math.Abs(sample + step) > volume)
                    step = -step;

                sample += step;
                samples[i] = sample;
            }

            return samples;
        }

        private double[] WaveWhiteNoise(int numSamples, double volume) {
            Random rng = new Random();
            double[] samples = new double[numSamples];
            for(int i = 0; i < numSamples; i++)
                samples[i] = ((rng.NextDouble() * 2) - 1) * volume;

            return samples;
        }
    }
}
